

/*************************	ECE 4760 FINAL PROJECT: AVINASH NAVADA (ABN44), KOFI OTSEIDU (KEO34)  ********************/


#define F_CPU 16000000UL

#include <util/delay.h>
#include <inttypes.h>
#include <avr/io.h> 
#include <avr/interrupt.h> 
#include <avr/pgmspace.h> 
#include <avr/eeprom.h> 
#include <avr/sleep.h>
#include <avr/power.h>
#include <stdio.h> 
#include <stdint.h>
#include <stdlib.h> 
#include <stddef.h>
#include <string.h> 
#include <math.h>
#include "uart.h"				// serial communication library	
#include "lcd_lib.h"			// LCD functions library
#include "i2c/i2cmaster.h"		// I2C Master library 
#include "sdcard/ff.h"			
#include "sdcard/diskio.h"			


// Camera defs
#define MAXLINES 292			// image height
#define MAXPIXELS 356			// image width
#define CAMADDR 0xc0
#define VSYNC		1
#define HREFUP		2
#define HREFDWN		3
#define HREF PORTC & (1<<PINC2) // PINC.2
#define VSYN PORTC & (1<<PINC3) // PINC.3
#define PCLK PORTD & (1<<PIND2) // PIND.2
#define IDLE		0
#define CAPTURE		1
#define SEND		2
// the usual
#define begin {
#define end }

FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

#define DEBUG		1
//prototypes
void turnOffCamera();

// global vars
volatile unsigned char is_alarm;					// alarm trigger signal
volatile unsigned int reset_count;
volatile unsigned char lineBuffer [MAXPIXELS];
volatile char cameraOn;
volatile char syncState;
volatile char commandFlag = IDLE;
volatile long pixelNum;
volatile long lineNum;
volatile long IR_counts;							// timer count
volatile long IR_period;							// # timer counts between IR pulses
volatile unsigned int period_found;					// IR pulse period found (training)
volatile unsigned int num_pulses;					// Max 2 (for training)
// edge detection
volatile unsigned int filter[] = {1,-2,1};
volatile unsigned int edges;						// number of edges in the image
volatile unsigned int sum;
volatile unsigned int max_edges;								// filter result per pixel
volatile unsigned char is_camera_finished;
//file system pointers
FIL file;
FIL* ptr; //= &logFile
FATFS fs;
FATFS* fsPtr;
unsigned char diskInitFlag;
unsigned char diskMountFlag;
unsigned char openFlag;
//time
volatile unsigned int millis;
volatile unsigned int secs;
volatile unsigned int mins;
volatile unsigned int hours;
volatile unsigned int day;
volatile unsigned int month;
volatile unsigned int year;
//char bt_buffer [10];								// Bluetooth data






/***********************************************************     CAMERA     ***********************************************************/

/****  Adapted from "Scanalicious: A 3D Scanner for You and Me", ECE 4760 Spring 2009 Project by Ryan Dunn and Dale Taylor  ******/


// I2C read from camera
unsigned char cameraRead(unsigned char regNum) 
{
	unsigned char write;
	unsigned char read;
	///Check these if they start to cause trouble
	write = i2c_start(CAMADDR+I2C_WRITE);   
	write = i2c_write(regNum);          
	i2c_stop();                              
	read = i2c_start(CAMADDR+I2C_READ); ; 
	read = i2c_readNak();
	i2c_stop();                              
	return read;
}

// I2C Write to camera
unsigned char cameraWrite(unsigned char regNum, unsigned char data) {
	unsigned char write;
	//Check these if they start to cause trouble
	write = i2c_start(CAMADDR+I2C_WRITE);   
	write = i2c_write(regNum);    
	write = i2c_write(data);          
	i2c_stop();   
	
	return 0;
}

void turnOnCamera()
{								
	//VREF - PINC3 interrupt set
	PCICR = (1<<PCIE2);
	PCMSK2 = (1<<PCINT19);	
	cameraOn = 1;	
	EIMSK &= ~(1<<INT0);			 				// turn off PCLK interrupt
	EICRA |= (1<<ISC01) | (1<<ISC00);  				// trigger interrupt on rising edge
	EIFR = 0x01;									// clear interrupt flag
}

///Ignore the camera's chatter
void turnOffCamera()
{
	///Turn off camera interrupts
	PCICR = 0x00;			// &= ~(1<<PCIE2);
	PCMSK2 = 0x00;			// &= ~(1<<PCINT19) & ~(1<<PCINT18);

	EIMSK &= ~(1<<INT0);							//turn off PCLK interrupt
	cameraOn = 0;
}

void initCam()
{

	/* 
	 * 16-BIT one line RGB-output mode
	 */

    _delay_ms(10);
    cameraWrite(0x12, 0x80);  // initiate a soft reset to make sure that all the regsiters are at default values
	_delay_ms(1000);
    cameraWrite(0x11, 0x2f);  //slow the clock way down
	_delay_ms(10);
    cameraWrite(0x12, 0x2c);  //COMMON CTRL A: RGB output
	_delay_ms(10);
    cameraWrite(0x28, 0x80);  //one line RGB output 
	_delay_ms(10);
    //cameraWrite(0x2a, 0xa0);  //Frame rate adjust on with slowest speed
	//_delay_ms(10);
	//cameraWrite(0x2b, 0xff);
	//_delay_ms(10);

	//Exposure
	cameraWrite(0x25, 0xff);
	_delay_ms(10);
	cameraWrite(0x24, 0x00);
	_delay_ms(10);
//	cameraWrite(0x2D, 0x04);
	_delay_ms(10);

	///Gain control
	//cameraWrite(0x01, 0x02);
	//_delay_ms(10);
	//cameraWrite(0x02, 0x02);
	//_delay_ms(10);

	//Saturation
	cameraWrite(0x03, 0x04);
	_delay_ms(10);

	//contrast/ brightness
	cameraWrite(0x05, 0xff);
	_delay_ms(10);
	cameraWrite(0x06,0x00);
	_delay_ms(10);
}

///Set up basic parameters for the camera
void setUpCamera()
{
	_delay_ms(10);
	cameraWrite(0x12, 0x80);  //soft reset
	_delay_ms(2);
	cameraWrite(0x11, 0x3f);  //clock div by 32
	_delay_ms(2);
	//cameraWrite(0x12, 0x2C);  //AGC enable, set RGB mode, with AWB enabled
	cameraWrite(0x12, 0x1C);  //8-bit digital out /w awb
	_delay_ms(2);
	cameraWrite(0x13, 0x21);  //tri-state the Y/UV lines, select 8 bit format
	_delay_ms(2);
    cameraWrite(0x28, 0x04);  // set color sequencer
    _delay_ms(2);
	cameraWrite(0x28, 0x80);  //one line RGB output  IS THIS RIGHT??
	_delay_ms(2);
	cameraWrite(0x2a, 0xa0);  //slow it down more
	_delay_ms(2);
	cameraWrite(0x2b, 0xff);
	_delay_ms(2);
	//		///TODO:  figure out what gain/white balance levels we need
}



/***********************************************************     BLUETOOTH    ***********************************************************/


/* Receive a data packet over UART from Bluetooth HC-05 module*/
/*void bt_read ( void )  
{
	unsigned int i = 0;
	char data;
	
	while(!(UCSR0A & (1<<RXC0)));	// wait until data is completely transmitted by HC-05
	
	data = UDR0;					// read char
	if (data != '#'){				// end of string
		bt_buffer[i] = data;
		i++;
	} 
	else{
		bt_buffer[i]='\0';
		i=0;
	}
}*/




/***********************************************************     SD CARD     ***********************************************************/



/* Inits filesystem on the SD card */
void init_fs() { 
	DSTATUS dstatus;
    
	dstatus = disk_initialize(0);
    fprintf(stdout, "\ndstatus: %d\n\r", dstatus);
    #ifdef DEBUG
	if(dstatus == STA_NOINIT) fprintf(stdout, "STA_NOINIT ");
	else if(dstatus == STA_NODISK) fprintf(stdout, "STA_NODISK ");
	else if(dstatus == STA_PROTECT) fprintf(stdout, "STA_PROTECT ");
    #endif
	
    if(dstatus == 0) diskInitFlag = 1; // Disk has been Initialized

    fprintf(stdout, "\nDisk Initialization: %d\n\r", diskInitFlag);
	FRESULT result;
	//init pointer
	fsPtr = &fs;
	ptr = &file;
	//set up file system
	result = f_mount(0, fsPtr);

    #ifdef DEBUG
	fprintf(stdout, "Initiating filesystem mount: %d\n\r", result);
    #endif
    
    if(result == 0) diskMountFlag = 1; // Disk has been mounted
}


/* write a line to the SD card */
void sd_write(char line[]) {
    FRESULT fileResult;
    unsigned int length;
    unsigned int bytesWrote;
    UINT bw;
	
	fileResult = f_write(ptr, line, sizeof(line), &bytesWrote);
	
	#ifdef DEBUG
	if (fileResult | (bytesWrote != length)) {
		fprintf(stdout, "Write error. Result %d, Tried %d, Wrote %d\r\n", fileResult, length, bytesWrote);
	}
	else {
		fprintf(stdout, "Write succeeded.\r\n");
	}
	#endif
	//sync the filesystem (flush buffers) so it will power down cleanly
	f_sync(ptr);
}


DWORD get_fattime()
{
	return      ((DWORD)(2012 - 1980) << 25)
            | ((DWORD)5 << 21)
            | ((DWORD)(day+1) << 16)
            | ((DWORD)hours << 11)
            | ((DWORD)mins << 5)
            | ((DWORD)(millis >> 10) >> 1); 
}





/***********************************************************     INTERRUPTS     ***********************************************************/



// Pclk (PIND2)
ISR(INT0_vect)
{ 			
	lineBuffer[pixelNum] = PINA;
	if (diskInitFlag && diskMountFlag) {
		sd_write(lineBuffer[pixelNum]);
	}

	sum = 0;									// filter result per pixel
	// Image Processing - Edge Detection
	for (int i=1; i<MAXPIXELS; i++) {
		sum = (lineBuffer[i-1]) - (lineBuffer[i]<<1) + (lineBuffer[i+1]);		// convolution with filter [1 -2  1]
		if (sum > 100) {
			edges++;
		}
	}
	//PORTD ^= 0x80;
	//fprintf(stdout, "%i\n", lineBuffer[0]);
	pixelNum++;
	//fprintf(stdout, "%ld\n", pixelNum);
	EIFR = 0x01;									// clear interrupt flag
}


/* Alarm (PIND3) */
ISR (INT1_vect)
{
	if (period_found == 0) {						// IR pulse period found (training)
		if (num_pulses < 2) {
			num_pulses++;
		}
	}
	// fprintf(stdout, "%ld", IR_counts);
	_delay_ms(27);
	IR_counts = 0;									// reset 
}


/* --- camera input ---*/
ISR (PCINT2_vect)
{ 	

	switch (syncState)
	{
		case VSYNC:

			syncState = HREFUP;
			lineNum = 0;							//reset line number
			PCMSK2 = (1<<PCINT18);					//turn on HREF ISR
			//fprintf(stdout,"\n\r VSYNC \n\r");
			break;
			
		case HREFUP:
			
			pixelNum = 0;
			syncState = HREFDWN;
			//turn on pixel clock
			EIMSK |= (1<<INT0);
			//fprintf(stdout,"\n\r HREFUP \n\r");
			break;

		case HREFDWN:	

			//increment line number
			lineNum++;
			//fprintf(stdout, "%ld\n", lineNum);
			//turn off pixel counter			
			EIMSK = 0x00;

			if(lineNum >= MAXLINES)
			{
				commandFlag = IDLE;
				syncState = VSYNC;
				PCMSK2 = 0x00;						//turn off href ISR
				turnOffCamera();
			    is_camera_finished = 1;
				// openFile an image and write it
				is_alarm = 0;
				EIMSK &= ~(1<<INT1);				// turn off alarm interrupt. PUT AT END OF BLUETOOTH FUNCTION
				EICRA |= (1<<ISC01);				// return trigger to negative edge
				EICRA &= ~(1<<ISC00);
				EIFR = 0x04;						// clear interrupt flag
				EIMSK |= 1<<INT1;					// turn on alarm interrupt
			}
	
			else
			{
				syncState = HREFUP;
				///Send some data
				commandFlag = SEND;
			}

			break;
			
	}
	//Clear the flag
	//PORTD = PORTD ^ (1<<PIND6);
	PCIFR = 0x00;
}






/***********************************************************     INITIALIZE     ***********************************************************/


void find_period ( void )
{
	EIMSK &= ~(1<<INT1); 				// turn off INT1
	EICRA |= (1<<ISC01);   				// trigger on falling edge
	EICRA &= ~(1<<ISC00);   
	EIFR = 0x04;						// clear interrupt flag
	EIMSK |= 1<<INT1; 					// turn on INT1

	int count = 0;

	// find interval between two consecutive pulses
	while (num_pulses < 2) {
		if (num_pulses == 1) {
			count++;
			_delay_ms(1);
		}

	} 
	
	IR_period = count;
	period_found = 1;
}


void initialize ( void ) 
{

	// Timer 2 Counter Setup
	//TCCR2B = 7;		  					// divide by 1024  
	//TIMSK2 = 1;		  					// timer 2 to be read in int0 ISR

	// IR Remote Receiver Setup
	DDRD &= ~0x08;						// setup PORTD.3 to read IR remote receiver
	PORTD |= 0x08;						// enable pull-up for D.3	    

	is_alarm = 0;						// alarm trigger signal
	period_found = 0;					
	num_pulses = 0;	

	edges = 0;							// number of edges in image

	// initialize the UART -- uart_init() is in uart.c
    uart_init();
    stdout = stdin = stderr = &uart_str;
    fprintf(stdout,"\n\r ECE 4760 Final Project: keo34, abn44 \n\r\n\r");

	// setup buzzer
	DDRD |= 0x10;						// setup Port D.4 for buzzer
	
	// set up i2c communcation
	i2c_init();

	// set up camera
	DDRD &= ~0x04;						// setup Port D.2 for PCLK
	PORTD |= 0x04;	  					// enable pull-up for D.2
	DDRA = 0x00;						// set PortA as an input (data)
	DDRC = DDRC & ~(1<< PORTC0);		// enable internal pull-ups for PortC 
	DDRC = DDRC & ~(1<< PORTC1);		
	PORTC = PORTC | (1<< PORTC0);		// C.0 - SCL
	PORTC = PORTC | (1<< PORTC1);		// C.1 - SDA
	DDRC = DDRC & ~(1 << PINC2) & ~(1 << PINC3); // set PCINTs 18 and 19 as inputs
	cameraOn = 0;
	initCam();
	commandFlag = IDLE;
	syncState = VSYNC;
	lineBuffer[0] = 0;			

	// set up SD Card ports
	DDRB = 0xbf;						// set PORTB to output except for B.6 (MISO)														
	PORTB = 0xff;	

	//init time
	millis = 0;
	secs = 0;
	mins = 0;
	hours = 0;
	day = 22; 							//dummy day
	month = 4; 							//dummy month
	year = 20; 							//dummy year

	// set up Bluetooth
/*	DDRD &= ~0x01;						// set D.0 (RX) to input
	DDRD |= 0x02;						// set D.1 (TX) to output
	PORTD |= 0x03;						// enable pull-ups for D.0 and D.1
	UCSR0A = 0x00;
	UCSR0B = 0x18;						// RXEN0, TXEN0 high
	UCSR0C = 0x06;						// UCSZ01, UCSZ00 high
	UBRR0H = 0x00;
	UBRR0L = 0x66;						// (16MHz / (16*9600 baud)) -1
	ACSR   = 0x80;*/

	// turn on all ISRs
	sei() ;		
	max_edges = 0;
	is_camera_finished = 0;
	DDRD |= 0x80;						// test LED
} 




/***********************************************************     MAIN     ***********************************************************/


int main(void) {


	initialize();

	find_period();
    fprintf(stdout, "IR_period: %ld\n", IR_period);


	while (1) {

		// alarm not triggered
		if (is_alarm == 0) {
			
			IR_counts++;
			_delay_ms(1);
			if (IR_counts > IR_period + 10)
			{
				// take a picture
				if(!cameraOn) {
					is_alarm = 1;
					EIMSK &= ~(1<<INT1);				// turn off alarm interrupt
				    is_camera_finished = 0;
					init_fs();							// initialize SD Card
					turnOnCamera();
				}
				
				// sound the alarm

				while(!is_camera_finished){}
				fprintf(stdout, "edges: %i\n", edges);
				if (edges > max_edges){
					max_edges = edges;
				}
				if(edges > 1200) {																						
					//PORTD |= 0x80;
					PORTD |= 0x10;
					_delay_ms(10000);
					PORTD &= ~0x10;
					_delay_ms(500);
					edges = 0;	
				}
				else if(edges == 0){								// avoid long delay if no image taken
					_delay_ms(1);
				}
				else{
					_delay_ms(2000);								// wait so continuous pictures are not taken
					edges = 0;
				}
                is_camera_finished = 0;
                IR_counts = 0;
                PCIFR = 0x00;										// clear all flags				
							
			}
			//start whole process over again


		}

	}

} 





